@@ -1,6 +1,7 @@
 from bz2 import BZ2File
 from collections import Counter, Sequence, Iterable, \
     Mapping
+import gc
 from email.mime.text import MIMEText
 import inspect
 from itertools import chain
@@ -232,6 +233,25 @@ def hsplit(text, sep, group=True, attach=True):
     return [word for word in chain(*zip(words, [sep]*len(words))) if word][:-1]
 
 
+def rmvars(*args):
+    """Wrapper to quickly free up memory by deleting global variables. Htools
+    3.0 does not provide a way to do this for local variables.
+
+    Parameters
+    ----------
+    args: str
+        One or more variable names to delete. Do not pass in the variable
+        itself.
+
+    Returns
+    -------
+    None
+    """
+    for arg in args:
+        del globals()[arg]
+    gc.collect()
+
+
 def print_object_sizes(space, limit=None, exclude_underscore=True):
     """Print the object names and sizes of the currently defined objects.
 
@@ -486,14 +506,20 @@ def select(items, keep=(), drop=()):
 
 
 def differences(obj1, obj2, methods=False, **kwargs):
-    """Find the differences between two objects of the same type. This is a
-    way to get more detail beyond whether two objects are equal or not.
+    """Find the differences between two objects (generally of the same type -
+    technically this isn't enforced but we do require that the objects have
+    the same set of attribute names so a similar effect is achieved. Actual
+    type checking was causing problems comparing multiple Args instances,
+    presumably because each Args object is defined when called).
+
+    This is a way to get more detail beyond whether two objects are equal or
+    not.
 
     Parameters
     -----------
-    obj1: any type
+    obj1: any
         An object.
-    obj2: same type as obj1
+    obj2: any, usually the same type as obj1
         An object.
     methods: bool
         If True, include methods in the comparison. If False, only attributes
@@ -525,10 +551,13 @@ def differences(obj1, obj2, methods=False, **kwargs):
         first is the corresponding value for obj1 and the second is the
         corresponding value for obj2.
     """
-    if obj1 == obj2:
-        return {}
+    # May built-in comparison functionality. Keep error handling broad.
+    try:
+        if obj1 == obj2:
+            return {}
+    except Exception:
+        pass
 
-    assert type(obj1) == type(obj2), 'Objects must be the same type.'
     attr1, attr2 = hdir(obj1, **kwargs), hdir(obj2, **kwargs)
     assert attr1.keys() == attr2.keys(), 'Objects must have same attributes.'
 
@@ -538,7 +567,7 @@ def differences(obj1, obj2, methods=False, **kwargs):
         if not (methods or v1 == 'attribute'):
             continue
 
-        # Comparisons work differently for numpy arrays.
+        # Comparisons work differently for arrays/tensors than other objects.
         val1, val2 = getattr(obj1, k1), getattr(obj2, k2)
         try:
             equal = (val1 == val2).all()
@@ -772,6 +801,71 @@ def item(it, random=True, try_values=True):
     return next(iter(it))
 
 
+def lmap(fn, *args):
+    """Basically a wrapper for `map` that returns a list rather than a
+    generator. This is such a common pattern that I think it deserves its own
+    function (think of it as a concise alternative to a list comprehension).
+    One slight difference is that we use *args instead of passing in an
+    iterable. This adds a slight convenience for the intended use case (fast
+    prototyping). See the `Examples` for more on this.
+
+    Parameters
+    ----------
+    args: any
+
+    Returns
+    -------
+    list
+
+    Examples
+    --------
+    Consider these three equivalent syntax options:
+
+    lmap(fn, x, y)
+    [fn(obj) for obj in (x, y)]
+    list(map(fn, (x, y))
+
+    When quickly iterating, option 1 saves a bit of typing. The extra
+    parentheses that options 2 and 3 require to put x and y in a temporary
+    data structure can get messy as we add more complex logic.
+    """
+    return list(map(fn, args))
+
+
+def attrmap(attr, *args):
+    """More convenient syntax for quick data exploration. Get an attribute
+    value for multiple objects.
+
+    Parameters
+    ----------
+    attr: str
+        Name of attribute to retrieve for each object.
+    args: any
+        Objects (usually of same type) to retrieve attributes for.
+
+    Returns
+    -------
+    list: Result for each object.
+
+    Examples
+    --------
+    df1 = pd.DataFrame(np.random.randint(0, 10, (4, 5)))
+    df2 = pd.DataFrame(np.random.randint(0, 3, (4, 5)))
+    df3 = pd.DataFrame(np.random.randint(0, 3, (2, 3)))
+
+    >>> attrmap('shape', df1, df2, df3)
+    [(4, 5), (4, 5), (2, 3)]
+
+    net = nn.Sequential(...)
+    >>> attrmap('shape', *net.parameters())
+    [torch.Size([5, 3]),
+     torch.Size([16, 4]),
+     torch.Size([16, 3]),
+     torch.Size([16])]
+    """
+    return [getattr(arg, attr) for arg in args]
+
+
 def identity(x):
     """Returns the input argument. Sometimes it is convenient to have this if
     we sometimes apply a function to an item: rather than defining a None
@@ -811,6 +905,111 @@ def max_key(d, fn=identity):
     return max(d.items(), key=lambda x: fn(x[1]))[0]
 
 
+def fgrep(text, term, window=25):
+    """Search a string for a given term. If found, print it with some context.
+    Similar to `grep -C 1 term text`. `fgrep` is short for faux grep.
+
+    Parameters
+    ----------
+    text: str
+        Text to search.
+    term: str
+        Term to look for in text.
+    window: int
+        Number of characters to display before and after the matching term.
+
+    Returns
+    -------
+    str: The desired term and its surrounding context. If the term isn't
+    present, an empty string is returned.
+    """
+    idx = text.find(term)
+    if idx == -1:
+        return ''
+    return text[max(idx-window, 0):idx+window]
+
+
+def spacer(char='-', n_chars=79, newlines_before=1, newlines_after=1):
+    """ Get string to separate output when printing output for multiple items.
+
+    Parameters
+    ----------
+    char: str
+        The character that will be printed repeatedly.
+    n_chars: int
+        The number of times to repeat `char`. We expect that `char` is a
+        single character so this will be the total line length.
+    newlines_before: int
+        Number of newline characters to add before the spacer.
+    newlines_after: int
+        Number of newline characters to add after the spacer.
+
+    Returns
+    -------
+    str
+    """
+    return '\n'*newlines_before + char * n_chars + '\n'*newlines_after
+
+
+def kwargs_fallback(self, *args, assign=False, **kwargs):
+    """Use inside a method that accepts **kwargs. Sometimes we want to use
+    an instance variable for some computation but want to give the user the
+    option to pass in a new value to the method (often ML hyperparameters) to
+    be used instead. This function makes that a little more convenient.
+
+    Parameters
+    ----------
+    self: object
+        The class instance. In most cases users will literally pass `self` in.
+    args: str
+        One or more names of variables to use this procedure on.
+    assign: bool
+        If True, any user-provided kwargs will be used to update attributes of
+        the instance. If False (the default), they will be used in computation
+        but won't change the state of the instance.
+    kwargs: any
+        Just forward along the kwargs passed to the method.
+
+    Returns
+    -------
+    list or single object: If more than one arg is specified, a list of values
+        is returned. For just one arg, a single value will be returned.
+
+    Examples
+    --------
+    class Foo:
+
+        def __init__(self, a, b=3, c=('a', 'b', 'c')):
+            self.a, self.b, self.c = a, b, c
+
+        def walk(self, d, **kwargs):
+            a, c = kwargs_fallback(self, 'a', 'c', **kwargs)
+            print(self.a, self.b, self.c)
+            print(a, c, end='\n\n')
+
+            b, c = kwargs_fallback(self, 'b', 'c', assign=True, **kwargs)
+            print(self.a, self.b, self.c)
+            print(b, c)
+
+    # Notice the first `kwargs_fallback` call doesn't change attributes of f
+    # but the second does. In the first block of print statements, the variable
+    # `b` does not exist yet because we didn't include it in *args.
+    >>> f = Foo(1)
+    >>> f.walk(d=0, b=10, c=100)
+    1 3 ('a', 'b', 'c')
+    1 100
+
+    1 10 100
+    10 100
+    """
+    res = []
+    for arg in args:
+        val = kwargs.get(arg) or getattr(self, arg)
+        res.append(val)
+        if assign: setattr(self, arg, val)
+    return res if len(res) > 1 else res[0]
+
+
 def cd_root(root_subdir='notebooks'):
     """Run at start of Jupyter notebook to enter project root.
 
