@@ -1271,11 +1271,10 @@ def wrapmethods(*decorators, methods=(), internals=False):
     return wrapper
 
 
-def delegate(attr):
+def delegate(attr, iter_magics=False, skip=(), getattr_=True):
     """Decorator that automatically delegates attribute calls to an attribute
     of the class. This is a nice convenience to have when using composition.
-    This does NOT affect magic methods; for that, see the `forwardable`
-    library.
+    User can also choose to delegate magic methods related to iterables.
 
     Note: I suspect this could lead to some unexpected behavior so be careful
     using this in production.
@@ -1284,6 +1283,18 @@ def delegate(attr):
     ----------
     attr: str
         Name of variable to delegate to.
+    iter_magics: bool
+        If True, delegate the standard magic methods related to iterables:
+        '__getitem__', '__setitem__', '__delitem__', and '__len__'.
+    skip: Iterable[str]
+        Can optionally provide a list of iter_magics to skip. This only has
+        an effect when `iter_magics` is True. For example, you may want to be
+        able to iterate over the class but no allow item deletion. In this case
+        you should pass skip=('__delitem__').
+    getattr_: bool
+        If True, delegate non-magic methods. This means that if you try to
+        access an attribute or method that the object produced by the decorated
+        class does not have, it will look for it in the delegated object.
 
     Examples
     --------
@@ -1299,9 +1310,9 @@ def delegate(attr):
     page = Page('http://www.coursera.org')
     page.find_all('div')
 
-    Example 2: Magic methods are not delegated.
+    Example 2: Magic methods except for __delitem__ are delegated.
 
-    @delegate('data')
+    @delegate('data', True, skip=('__delitem__'))
     class Foo:
         def __init__(self, data, city):
             self.data = data
@@ -1309,15 +1320,62 @@ def delegate(attr):
 
     >>> f = Foo(['a', 'b', 'c'], 'San Francisco')
     >>> len(f)
+    3
 
-    TypeError: object of type 'Foo' has no len()
+    >>> for char in f:
+    >>>     print(char)
+    a
+    b
+    c
+
+    >>> f.append(3); f.data
+    ['a', 'b', 'c', 3]
+
+    >>> del f[0]
+    TypeError: 'Foo' object doesn't support item deletion
+
+    >>> f.clear(); f.data
+    []
     """
     def wrapper(cls):
-        def f(self, new_attr):
-            delegate = getattr(self, attr)
-            return getattr(delegate, new_attr)
-        cls.__getattr__ = f
+        def _delegate(self, attr):
+            """Helper that retrieves object that an instance delegates to."""
+            return getattr(self, attr)
+
+        # Any missing attribute will be delegated.
+        if getattr_:
+            def _getattr(self, new_attr):
+                return getattr(_delegate(self, attr), new_attr)
+
+            cls.__getattr__ = _getattr
+
+        # If specified, delegate magic methods to make cls iterable.
+        if iter_magics:
+            if '__getitem__' not in skip:
+                def _getitem(self, i):
+                    return _delegate(self, attr)[i]
+
+                setattr(cls, '__getitem__', _getitem)
+
+            if '__setitem__' not in skip:
+                def _setitem(self, i, val):
+                    _delegate(self, attr)[i] = val
+
+                setattr(cls, '__setitem__', _setitem)
+
+            if '__delitem__' not in skip:
+                def _delitem(self, i):
+                    del _delegate(self, attr)[i]
+
+                setattr(cls, '__delitem__', _delitem)
+
+            if '__len__' not in skip:
+                def _len(self):
+                    return len(_delegate(self, attr))
+
+                setattr(cls, '__len__', _len)
         return cls
+
     return wrapper
 
 
@@ -1440,7 +1498,8 @@ def copy_func(func):
     new_func = types.FunctionType(func.__code__, func.__globals__,
                                   func.__name__, func.__defaults__,
                                   func.__closure__)
-    new_func.__kwdefaults__ = func.__kwdefaults__.copy()
+    defaults = getattr(func, '__kwdefaults__') or {}
+    new_func.__kwdefaults__ = defaults.copy()
     return update_wrapper(new_func, func)
 
 
@@ -1473,14 +1532,36 @@ def rename_params(func, **old2new):
     now be named "y_true" and "y_pred", respectively. """
     new_func = copy_func(func)
     sig = inspect.signature(new_func)
-    kw_defaults = func.__kwdefaults__
+    kw_defaults = func.__kwdefaults__ or {}
     names, params = map(list, zip(*sig.parameters.items()))
     for old, new in old2new.items():
         idx = names.index(old)
-        default = kw_defaults.get(old)
-        kwargs = {'default': default} if default else {}
-        params[idx] = inspect.Parameter(new, params[idx].kind, **kwargs)
+        default = kw_defaults.get(old) or params[idx].default
+        params[idx] = inspect.Parameter(new, params[idx].kind, default=default)
     new_func.__signature__ = sig.replace(parameters=params)
     return new_func
 
 
+def immutify_defaults(func):
+    """Decorator to make a function's defaults arguments effectively immutable.
+    We accomplish this by storing the initially provided defaults and assigning
+    them back to the function's signature after each call. If you use a
+    variable as a default argument, this does not mean that the variable's
+    value will remain unchanged - it just ensures the initially provided value
+    will be used for each call.
+    """
+    # If `__hash__` is not None, object is immutable already.
+    # Python sets __defaults__ and __kwdefaults__ to None when they're empty.
+    _defaults = tuple(o if getattr(o, '__hash__') else deepcopy(o)
+                      for o in getattr(func, '__defaults__') or ()) or None
+    _kwdefaults = {k: v if getattr(v, '__hash__') else deepcopy(v) for k, v
+                   in (getattr(func, '__kwdefaults__') or {}).items()} or None
+
+    @wraps(func)
+    def wrapper(*args, **kwargs):
+        res = func(*args, **kwargs)
+        wrapper.__defaults__ = func.__defaults__ = deepcopy(_defaults)
+        wrapper.__kwdefaults__ = func.__kwdefaults__ = deepcopy(_kwdefaults)
+        return res
+    return wrapper
+
